/* script.c
 *
 * Functions to call the DHCP client notification scripts 
 *
 * Russ Dill <Russ.Dill@asu.edu> July 2001
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/fcntl.h>
#include <sys/wait.h>
#include <errno.h>

#include "options.h"
#include "dhcpd.h"
#include "dhcpc.h"
#include "packet.h"
#include "options.h"
#include "debug.h"

#include "../shared/include/nvram.h"

extern int unit;
extern char str_wan_ip[20];
extern char str_wan_mask[20];
extern char str_wan_gateway[20];
extern char str_wanif[10];

#define DPRINTF(args...)

//static int
//udhcpc_show_packet(struct dhcpMessage *packet);

static int
udhcpc_write_status_file(struct dhcpMessage *, const char *);


/* get a rough idea of how long an option will be (rounding up...) */
static int max_option_length[] = {
	[OPTION_IP] =		sizeof("255.255.255.255 "),
	[OPTION_IP_PAIR] =	sizeof("255.255.255.255 ") * 2,
	[OPTION_STRING] =	1,
	[OPTION_BOOLEAN] =	sizeof("yes "),
	[OPTION_U8] =		sizeof("255 "),
	[OPTION_U16] =		sizeof("65535 "),
	[OPTION_S16] =		sizeof("-32768 "),
	[OPTION_U32] =		sizeof("4294967295 "),
	[OPTION_S32] =		sizeof("-2147483684 "),
};


static int upper_length(int length, struct dhcp_option *option)
{
	return max_option_length[option->flags & TYPE_MASK] *
	       (length / option_lengths[option->flags & TYPE_MASK]);
}


static int sprintip(char *dest, char *pre, unsigned char *ip) {
//	return sprintf(dest, "%s%d.%d.%d.%d ", pre, ip[0], ip[1], ip[2], ip[3]);
	return sprintf(dest, "%s%d.%d.%d.%d ", pre, ip[0], ip[1], ip[2], ip[3]);
}
static int sprintip1(char *dest, char *pre, unsigned char *ip) {
//	return sprintf(dest, "%s%d.%d.%d.%d ", pre, ip[0], ip[1], ip[2], ip[3]);
	return sprintf(dest, "%s%d.%d.%d.%d", pre, ip[0], ip[1], ip[2], ip[3]);
}

/* Fill dest with the text of option 'option'. */
static void fill_options(char *dest, unsigned char *option, struct dhcp_option *type_p)
{
	int type, optlen;
	u_int16_t val_u16;
	int16_t val_s16;
	u_int32_t val_u32;
	int32_t val_s32;
	int len = option[OPT_LEN - 2];

	dest += sprintf(dest, "%s=", type_p->name);

	type = type_p->flags & TYPE_MASK;
	optlen = option_lengths[type];
	for(;;) {
		switch (type) {
		case OPTION_IP_PAIR:
			dest += sprintip(dest, "", option);
			*(dest++) = '/';
			option += 4;
			optlen = 4;
		case OPTION_IP:	/* Works regardless of host byte order. */
			dest += sprintip(dest, "", option);
 			break;
		case OPTION_BOOLEAN:
			dest += sprintf(dest, *option ? "yes " : "no ");
			break;
		case OPTION_U8:
			dest += sprintf(dest, "%u ", *option);
			break;
		case OPTION_U16:
			memcpy(&val_u16, option, 2);
			dest += sprintf(dest, "%u ", ntohs(val_u16));
			break;
		case OPTION_S16:
			memcpy(&val_s16, option, 2);
			dest += sprintf(dest, "%d ", ntohs(val_s16));
			break;
		case OPTION_U32:
			memcpy(&val_u32, option, 4);
			dest += sprintf(dest, "%lu ", (unsigned long) ntohl(val_u32));
			break;
		case OPTION_S32:
			memcpy(&val_s32, option, 4);
			dest += sprintf(dest, "%ld ", (long) ntohl(val_s32));
			break;
		case OPTION_STRING:
			memcpy(dest, option, len);
			dest[len] = '\0';
			return;	 /* Short circuit this case */
		}
		option += optlen;
		len -= optlen;
		if (len <= 0) break;
	}
}


static char *find_env(const char *prefix, char *defaultstr)
{
	extern char **environ;
	char **ptr;
	const int len = strlen(prefix);

	for (ptr = environ; *ptr != NULL; ptr++) {
		if (strncmp(prefix, *ptr, len) == 0)
			return *ptr;
	}
	return defaultstr;
}


/* put all the paramaters into an environment */
static char **fill_envp(struct dhcpMessage *packet)
{
	int num_options = 0;
	int i, j;
	char **envp;
	unsigned char *temp;
	char over = 0;

	if (packet == NULL)
		num_options = 0;
	else {
		for (i = 0; options[i].code; i++)
			if (get_option(packet, options[i].code))
				num_options++;
		if (packet->siaddr) num_options++;
		if ((temp = get_option(packet, DHCP_OPTION_OVER)))
			over = *temp;
		if (!(over & FILE_FIELD) && packet->file[0]) num_options++;
		if (!(over & SNAME_FIELD) && packet->sname[0]) num_options++;		
	}

	envp = malloc((num_options + 6) * sizeof(char *));
	envp[0] = malloc(sizeof("interface=") + strlen(client_config.interface));
	sprintf(envp[0], "interface=%s", client_config.interface);
	envp[1] = find_env("PATH", "PATH=/bin:/usr/bin:/sbin:/usr/sbin");
	envp[2] = find_env("HOME", "HOME=/");

	if (packet == NULL) {
		envp[3] = NULL;
		return envp;
	}

	envp[3] = malloc(sizeof("ip=255.255.255.255"));
	sprintip(envp[3], "ip=", (unsigned char *) &packet->yiaddr);
	for (i = 0, j = 4; options[i].code; i++) {
		if ((temp = get_option(packet, options[i].code))) {
			envp[j] = malloc(upper_length(temp[OPT_LEN - 2], &options[i]) + strlen(options[i].name) + 2);
			fill_options(envp[j], temp, &options[i]);
			j++;
		}
	}
	if (packet->siaddr) {
		envp[j] = malloc(sizeof("siaddr=255.255.255.255"));
		sprintip(envp[j++], "siaddr=", (unsigned char *) &packet->yiaddr);
	}
	if (!(over & FILE_FIELD) && packet->file[0]) {
		/* watch out for invalid packets */
		packet->file[sizeof(packet->file) - 1] = '\0';
		envp[j] = malloc(sizeof("boot_file=") + strlen(packet->file));
		sprintf(envp[j++], "boot_file=%s", packet->file);
	}
	if (!(over & SNAME_FIELD) && packet->sname[0]) {
		/* watch out for invalid packets */
		packet->sname[sizeof(packet->sname) - 1] = '\0';
		envp[j] = malloc(sizeof("sname=") + strlen(packet->sname));
		sprintf(envp[j++], "sname=%s", packet->sname);
	}
	envp[j] = malloc(sizeof("unit=") + 5);
	sprintf(envp[j++], "unit=%d", unit);
	envp[j] = NULL;
	return envp;
}


/* Call a script with a par file and env vars */
void run_script1(struct dhcpMessage *packet, const char *name)
{
	int pid;
	char **envp;
	int is_wan_ip_the_same=0;

	if (client_config.script == NULL)
		return;

        is_wan_ip_the_same=udhcpc_write_status_file(packet, name);
        if(!strcmp(name,"deconfig") || !strcmp(name,"bound")){
            if (is_wan_ip_the_same==1&&!strcmp(name,"bound")){
                LOG(LOG_INFO, "the renew ip is same with previous ip");
		return;                                
            }
        }


	/* call script */
	pid = vfork();
	if (pid) {
		waitpid(pid, NULL, 0);
		return;
	} else if (pid == 0) {
		envp = fill_envp(packet);
		
		/* close fd's? */
		
		/* exec script */
		//DEBUG(LOG_INFO, "execle'ing %s", client_config.script);
		//DPRINTF("execle'ing %s\n", client_config.script);
		DEBUG(LOG_INFO, "executing .....");
		execle(client_config.script, client_config.script,
		       name, NULL, envp);
		//LOG(LOG_ERR, "script %s failed: %s",
		//  client_config.script, sys_errlist[errno]);
		LOG(LOG_ERR, "executing failed ..");
		exit(1);
	}		
}



/* the two files for showing home.asp and running fw.sh */
static int
udhcpc_write_status_file(struct dhcpMessage *packet, const char *name)
{
  int optlen, type;
  int num_options = 0;
  int i,DNS_No,kj;
  unsigned char *temp, *str;
  char over = 0;
  char str_dns[20];
  char buf[64];
  FILE *fp;

  DPRINTF("run in %s\n", name);
  if(strcmp(name, "deconfig")==0){
    UpdateWANIP(unit, client_config.interface, "", "", ""); //wan_if, ip, mask, gw
  }
  if( strcmp(name, "renew")!=0 && strcmp(name, "bound")!=0)
    return 0;
  else{	

	
  if (packet == NULL)
    num_options = 0;
  else {
    for (i = 0; options[i].code; i++)
      if (get_option(packet, options[i].code))
	num_options++;
    if (packet->siaddr) num_options++;
    if ((temp = get_option(packet, DHCP_OPTION_OVER)))
      over = *temp;
    if (!(over & FILE_FIELD) && packet->file[0]) num_options++;
    if (!(over & SNAME_FIELD) && packet->sname[0]) num_options++;		
  }

  if (packet == NULL) {
    return 0;
  }

	
  /* WAN_IF */
  sprintf(str_wanif, client_config.interface);
  DPRINTF("interface=%s\n", str_wanif);

  /* WAN_IP */
//  sprintip(str_wan_ip, "", (unsigned char *) &packet->yiaddr);
  sprintip1(str_wan_ip, "", (unsigned char *) &packet->yiaddr);
  DPRINTF("yiaddr=%s\n", str_wan_ip);

  
  for (i = 0; options[i].code; i++) {
    if ((temp = get_option(packet, options[i].code))) {
      //DPRINTF("options name=%s\n", options[i].name);
      /* WAN_MASK */
      if(strcmp(options[i].name, "subnet")==0){
	//sprintip(str_wan_mask, "", temp);
	sprintip1(str_wan_mask, "", temp);
	DPRINTF("enter wan mask %s\n", str_wan_mask);
      }
      /* GATEWAY */
      if(strcmp(options[i].name, "router")==0){
	//sprintip(str_wan_gateway, "", temp);
	sprintip1(str_wan_gateway, "", temp);
	DPRINTF("enter gateway %s\n", str_wan_gateway);
      }
      /* DNS */
      
      if(strcmp(options[i].name, "dns")==0) {
	type = options[i].flags & TYPE_MASK;
	optlen = option_lengths[type];
//	DPRINTF("optlen=\"%d\"\n", optlen);
	
	//It is possible that we have more than 2 dns .
	// So, we have to look up the length of DNS IP address
	DNS_No=(*(temp-1))/4;
	for(kj=0;kj<DNS_No;kj++)
	{
		sprintip(str_dns, "", temp);
		DPRINTF("dns #%d: %s\n",kj,str_dns);
		temp+=4;
	}	
     }
    }
  }

}
  return 0;

}

