/*
** $Id: sha1.c 1.3 2004/05/04 17:13:56 chyeh Exp $
*/

/*************************************************************************
* Copyright (c) 2002-2003 Inprocomm, Inc.
*
* All rights reserved. Copying, compilation, modification, distribution
* or any other use whatsoever of this material is strictly prohibited
* except in accordance with a Software License Agreement with
* Inprocomm, Inc.
**************************************************************************/
/*************************************************************************
GENERAL DESCRIPTION
********************
1. This file implement SHA-1 Secure Hash Algorithm. The SHA-1 is
   implemented according to FIPS 180-1
2. This file implement Keyed-Hashing for Message Authentication: HMAC-SHA1.
   The HMAC-SHA1 is implemented according to RFC2104 (Keyed-Hashing for
   Message Authentication, HMAC)
3. This file implement Pseudo-Random Function, PRF. The PRF is implemented
   according to IEEE 802.11i on page 157.
*/

/************************************************************************
*                E X T E R N A L   R E F E R E N C E S
*************************************************************************
*/
#include <linux/string.h>
#include "precomp.h"

/************************************************************************
*                          C O N S T A N T S
*************************************************************************
*/

/************************************************************************
*                         D A T A   T Y P E S
*************************************************************************
*/
//typedef struct {
//    ULONG bitCount[2];     /* processed bit count */
//    ULONG unProcByteCount; /* number of un-processed byte count in X[] */
//    ULONG H[5];
//    ULONG W[80];
//    ULONG K[80];
//    ULONG X[16];
//    ULONG TEMP;
//    ULONG A, B, C, D, E;
//} SHA1_CTX, *PSHA1_CTX;

/************************************************************************
*                        P U B L I C   D A T A
*************************************************************************
*/

/************************************************************************
*                       P R I V A T E   D A T A
*************************************************************************
*/

/************************************************************************
*                             M A C R O S
*************************************************************************
*/
#define S(n, X) (((X) << (n)) | ((X) >> (32 - (n))))

/************************************************************************
*              F U N C T I O N   D E C L A R A T I O N S
*************************************************************************
*/

/************************************************************************
*                          F U N C T I O N S
*************************************************************************
*/
ULONG
f(IN ULONG t,
  IN ULONG B,
  IN ULONG C,
  IN ULONG D)
{
    if (t <= 19) {
        return ((B & C) | ((~B) & D));
    }
    else if ((t >= 20) && (t <= 39)) {
        return (B ^ C ^ D);
    }
    else if ((t >= 40) && (t <= 59)) {
        return ((B & C) | (B & D) | (C & D));
    }
    else if ((t >= 60) && (t <= 79)) {
        return (B ^ C ^ D);
    }
    else {
        return 0;
    }
}

/***************************************************************************
* Function   : sha1Encode
* Input      : input_p, pointer to 32-bit value
*              outputLen, length of output string, in bytes
* Output     : output_p, pointer to starting address of first 8-bit value
* Return     :
* Description: This function encode single 32-bit value into 4 8-bit values
* Note       :
****************************************************************************/
VOID
sha1Encode(IN PULONG input_p,
           OUT PUCHAR output_p,
           IN ULONG outputLen)
{
    ULONG i, j;
    for (i = 0, j = 0; j < outputLen; i++, j += 4){
        output_p[j    ] = (UCHAR)((input_p[i] >> 24) & 0x000000ff);
        output_p[j + 1] = (UCHAR)((input_p[i] >> 16) & 0x000000ff);
        output_p[j + 2] = (UCHAR)((input_p[i] >>  8) & 0x000000ff);
        output_p[j + 3] = (UCHAR)(input_p[i]         & 0x000000ff);
    }
}

/***************************************************************************
* Function   : sha1Decode
* Input      : input_p, pointer to starting address of first 8-bit value
*              inputLen, length of output string, in bytes
* Output     : output_p, pointer to 32-bit value
* Return     :
* Description: This function decode 4 8-bit values into single 32-bit value
* Note       :
****************************************************************************/
VOID
sha1Decode(IN PUCHAR input_p,
           IN ULONG inputLen,
           OUT PULONG output_p)
{
    ULONG i, j;
    for (i = 0, j = 0; j < inputLen; i++, j += 4){
        output_p[i] = (input_p[j  ] << 24) |
                      (input_p[j+1] << 16) |
                      (input_p[j+2] <<  8) |
                       input_p[j+3];
    }
}

/***************************************************************************
* Function   : sha1Init
* Input      : ctx_p, pointer to SHA-1 context
* Output     :
* Return     :
* Description: This function initialize SHA-1 context value
* Note       :
****************************************************************************/
VOID
sha1Init(IN PSHA1_CTX ctx_p)
{
    ULONG i;

    /* initialize value */
    for (i = 0; i < 20; i++) {
        ctx_p->K[   i] = 0x5A827999;
        ctx_p->K[20+i] = 0x6ED9EBA1;
        ctx_p->K[40+i] = 0x8F1BBCDC;
        ctx_p->K[60+i] = 0xCA62C1D6;
        ctx_p->W[   i] = 0;
        ctx_p->W[20+i] = 0;
        ctx_p->W[40+i] = 0;
        ctx_p->W[60+i] = 0;
    }

    ctx_p->H[0] = 0x67452301;
    ctx_p->H[1] = 0xEFCDAB89;
    ctx_p->H[2] = 0x98BADCFE;
    ctx_p->H[3] = 0x10325476;
    ctx_p->H[4] = 0xC3D2E1F0;

    ctx_p->unProcByteCount = 0;
    ctx_p->bitCount[0] = 0;
    ctx_p->bitCount[1] = 0;
    ctx_p->TEMP = 0;
    ctx_p->A = 0;
    ctx_p->B = 0;
    ctx_p->C = 0;
    ctx_p->D = 0;
    ctx_p->E = 0;
}

/***************************************************************************
* Function   : sha1Transform
* Input      : ctx_p, pointer to SHA-1 context
* Output     :
* Return     :
* Description: This function perform SHA-1 algorithm with specified
*              SHA-1 context
* Note       :
****************************************************************************/
VOID
sha1Transform(IN PSHA1_CTX ctx_p)
{
    ULONG i;

    sha1Decode((PUCHAR )&ctx_p->X[0], 64, &ctx_p->X[0]);

    /* 7.a */
    for (i = 0; i < 16; i++){
        ctx_p->W[i] = ctx_p->X[i];
    }

    /* 7.b */
    for (i = 16; i < 80; i++){
        ctx_p->W[i] = S(1, ctx_p->W[i-3] ^ ctx_p->W[i-8] ^ ctx_p->W[i-14] ^ ctx_p->W[i-16]);
    }

    /* 7.c */
    ctx_p->A = ctx_p->H[0];
    ctx_p->B = ctx_p->H[1];
    ctx_p->C = ctx_p->H[2];
    ctx_p->D = ctx_p->H[3];
    ctx_p->E = ctx_p->H[4];

    /* 7.d */
    for (i = 0; i < 80; i++){
        ctx_p->TEMP = S(5, ctx_p->A) + f(i, ctx_p->B, ctx_p->C, ctx_p->D) + ctx_p->E + ctx_p->W[i] + ctx_p->K[i];
        ctx_p->E = ctx_p->D;
        ctx_p->D = ctx_p->C;
        ctx_p->C = S(30, ctx_p->B);
        ctx_p->B = ctx_p->A;
        ctx_p->A = ctx_p->TEMP;
    }

    /* 7.e */
    ctx_p->H[0] += ctx_p->A;
    ctx_p->H[1] += ctx_p->B;
    ctx_p->H[2] += ctx_p->C;
    ctx_p->H[3] += ctx_p->D;
    ctx_p->H[4] += ctx_p->E;
}

/***************************************************************************
* Function   : sha1Update
* Input      : ctx_p, pointer to SHA-1 context
*              input_p, pointer to input data
*              inputLen, length of input data
* Output     :
* Return     :
* Description: This function perform SHA-1 algorithm with specified SHA-1
*              context and first 1 ~ N-1 data block
* Note       : The last data block N is put in X of context
****************************************************************************/
VOID
sha1Update(IN PSHA1_CTX ctx_p,
           IN PUCHAR input_p,
           IN ULONG inputLen)
{
    ULONG i;
    ULONG remainLen;
    ULONG curInputIdx;

    remainLen = inputLen;
    curInputIdx = 0;

    if (ctx_p->unProcByteCount + inputLen >= 64) {
        while (ctx_p->unProcByteCount + remainLen >= 64) {
            for (i = ctx_p->unProcByteCount; i < 64; i++) {
                *((PUCHAR )&ctx_p->X[0]+i) = input_p[curInputIdx];
                curInputIdx++;
            }
            remainLen -= (64 - ctx_p->unProcByteCount);
            sha1Transform(ctx_p);
            ctx_p->unProcByteCount = 0;
        }
    }

    if (0xffffffff - ctx_p->bitCount[0] < (ULONG)inputLen * 8) {
        ctx_p->bitCount[1]++;
    }
    ctx_p->bitCount[0] += inputLen * 8;

    /* clear X[] */
    for (i = 0; i < 16; i++) {
        ctx_p->X[i] = 0;
    }

    if (remainLen == 0) {
        ctx_p->unProcByteCount = 0;
    }
    else {
        ctx_p->unProcByteCount = remainLen;
        /* copy remain input data to ctx_p->X[] */
        for (i = 0; i < remainLen; i++) {
            *((PUCHAR )&ctx_p->X[0] + i) = input_p[curInputIdx + i];
        }
    }
}

/***************************************************************************
* Function   : sha1Final
* Input      : ctx_p, pointer to SHA-1 context
* Output     : digest_p, pointer to the address to put the digest
* Return     :
* Description: This function perform SHA-1 algorithm with specified SHA-1
*              context and last data block N. Finally, this function put
*              the digest into specified digest_p.
* Note       :
****************************************************************************/
VOID
sha1Final(IN PSHA1_CTX ctx_p,
          OUT PUCHAR digest_p)
{
    ULONG i;
    /* Process last block:
       1. padding 1
       2. padding 0's
       3. padding input data length, in bits
       4. SHA1 final transform
     */

    /* 1. padding 1 */
    *((PUCHAR )&ctx_p->X[0] + ctx_p->unProcByteCount) = 0x80;

    /* 2. padding 0's. Note: we have pdded 0's in ctx.W[] in sha1Update() */
    if (64 - ctx_p->unProcByteCount <= 8) {
        sha1Transform(ctx_p);
        for (i=0; i<16; i++){
            ctx_p->X[i] = 0;
        }
    }

    /* 3. padding input data length, in bits */
    sha1Encode(&ctx_p->bitCount[0], (PUCHAR )&ctx_p->X[15], 4);

    /* 4. SHA1 final transform */
    sha1Transform(ctx_p);
    sha1Encode(ctx_p->H, digest_p, 20);
}

/***************************************************************************
* Function   : hmacSha1
* Input      : data_p, pointer to input data
*              dataLen, length of input data, in bytes
*              key_p, pointer to key
*              keyLen, length of key, in bytes
* Output     : digest_p, pointer to the address to put the digest
* Return     :
* Description: This function perform Keyed-Hashing for Message
*              Authentication (HMAC) using SHA-1 algorithm. Finally, this
*              function put the digest into specified digest_p.
* Note       :
****************************************************************************/
VOID
hmacSha1(IN PUCHAR data_p,
         IN ULONG dataLen,
         IN PUCHAR key_p,
         IN ULONG keyLen,
         OUT PUCHAR digest_p)
{
    SHA1_CTX context;
    UCHAR k_ipad[65];
    UCHAR k_opad[65];
    ULONG i;

    if (keyLen > 64) {
        SHA1_CTX tctx;

        sha1Init(&tctx);
        sha1Update(&tctx, key_p, keyLen);
        sha1Final(&tctx, key_p);

        keyLen = 20;
    }

    for (i = 0; i < 65; i++) {
        k_ipad[i] = 0;
        k_opad[i] = 0;
    }

    for (i = 0; i < keyLen; i++) {
        k_ipad[i] = key_p[i];
        k_opad[i] = key_p[i];
    }

    for (i = 0; i < 64; i++) {
        k_ipad[i] ^= 0x36;
        k_opad[i] ^= 0x5c;
    }

    sha1Init(&context);
    sha1Update(&context, k_ipad, 64);
    sha1Update(&context, data_p, dataLen);
    sha1Final(&context, digest_p);

    sha1Init(&context);
    sha1Update(&context, k_opad, 64);
    sha1Update(&context, digest_p, 20);
    sha1Final(&context, digest_p);
}

/***************************************************************************
* Function   : prf
* Input      : key_p, pointer to key
*              keyLen, length of key, in bytes
*              prefix_p, pointer to prefix
*              prefixLen, length of prefix, in bytes
*              data_p, pointer to input data
*              dataLen, length of input data, in bytes
*              output_p, pointer to output string
*              outputLen, length of output string
*
* Output     : digest_p, pointer to the address to put the digest
* Return     :
* Description: This function perform Pseudo-Random Function
* Note       :
****************************************************************************/
VOID
prf(IN PUCHAR key_p,
    IN ULONG keyLen,
    IN PUCHAR prefix_p,
    IN ULONG prefixLen,
    IN PUCHAR data_p,
    IN ULONG dataLen,
    OUT PUCHAR output_p,
    IN ULONG outputLen)
{
    ULONG i;
    UCHAR input_p[1024];
    ULONG currentIndex = 0;
    ULONG totalLen;

    for (i = 0; i < prefixLen; i++) {
        input_p[i] = prefix_p[i];
    }
    input_p[prefixLen] = 0;
    for (i = 0; i < dataLen; i++) {
        input_p[prefixLen + 1 + i] = data_p[i];
    }
    totalLen = prefixLen + 1 + dataLen;
    input_p[totalLen] = 0;
    totalLen++;

    for (i = 0; i < (outputLen + 19)/20; i++) {
        hmacSha1(input_p, totalLen, key_p, keyLen, &output_p[currentIndex]);
        currentIndex += 20;
        input_p[totalLen - 1]++;
    }
}

/***************************************************************************
* Function   : F
* Input      : password_p, pointer to pass-phrase password string
*              ssid_p, pointer to ssid string
*              ssidLen, length of ssid string
*              iterations, number of times the pass-phrase to hashed
*              count,
*
* Output     : output_p, pointer to the address to put the result
* Return     :
* Description: F(P, S, c, i) = U1 xor U2 xor ... Uc
*              U1 = PRF(P, S || Int(i))
*              U2 = PRF(P, U1)
*              Uc = PRF(P, Uc-1)
* Note       :
****************************************************************************/
VOID
F(IN PCHAR password_p,
  IN PUCHAR ssid_p,
  IN INT ssidLen,
  IN INT iterations,
  IN INT count,
  OUT PUCHAR output_p)
{
    UCHAR digest[36], digest1[32];
    INT i, j;

    for (i = 0; i < 36; i++) {
        digest[i] = 0;
    }

    for (i = 0; i < 32; i++) {
        digest1[i] = 0;
    }

    /* U1 = PRF(P, S || INT(i)) */
    memcpy(digest, ssid_p, ssidLen);
    digest[ssidLen    ] = (UCHAR)((count >> 24) & 0xff);
    digest[ssidLen + 1] = (UCHAR)((count >> 16) & 0xff);
    digest[ssidLen + 2] = (UCHAR)((count >>  8) & 0xff);
    digest[ssidLen + 3] = (UCHAR)(count         & 0xff);
    hmacSha1(digest, ssidLen + 4, (PUCHAR) password_p,
        (INT) strlen(password_p), digest1);

    /* output = U1 */
    memcpy(output_p, digest1, 20);

    for (i = 1; i < iterations; i++) {
        /* Un = PRF(P, Un-1) */
        hmacSha1(digest1, 20, (PUCHAR) password_p,
            (INT) strlen(password_p), digest);
        memcpy(digest1, digest, 20);

        /* output = output xor Un */
        for (j = 0; j < 20; j++) {
            output_p[j] ^= digest[j];
        }
    }
}

/***************************************************************************
* Function   : passwordHash
* Input      : password_p, pointer to pass-phrase password string
*              ssid_p, pointer to ssid string
*              ssidLen, length of ssid string
*
* Output     : output_p, pointer to the address to put the result
* Return     :
* Description: password - ascii string up to 63 characters in length
*              ssid - octet string up to 32 octets
*              ssidLen - length of ssid in octets
*              output must be 40 octets in length and outputs 256 bits of key
* Note       :
****************************************************************************/
VOID
passwordHash(IN PCHAR password_p,
             IN PUCHAR ssid_p,
             IN INT ssidLen,
             OUT PUCHAR output_p)
{
    F(password_p, ssid_p, ssidLen, 4096, 1, output_p);
    F(password_p, ssid_p, ssidLen, 4096, 2, &output_p[20]);
}

/***************************************************************************
* Function   : passwordHash
* Input      : password_p, pointer to pass-phrase password string
*              ssid_p, pointer to ssid string
*              ssidLen, length of ssid string
*              iterations, number of times the pass-phrase to hashed
*              outputLen, length of output key, in bytes
*
* Output     : output_p, pointer to the address to put the result
* Return     :
* Description: password - ascii string up to 63 characters in length
*              ssid - octet string up to 32 octets
*              ssidLen - length of ssid in octets
*              output must be 40 octets in length and outputs 256 bits of key
* Note       :
****************************************************************************/
BOOLEAN
pbkdf2(IN PCHAR password_p,
       IN PUCHAR ssid_p,
       IN INT ssidLen,
       IN INT iteration,
       IN INT outputLen,
       OUT PUCHAR output_p)
{
    UCHAR keyText[40];

    if ((strlen(password_p) > 63) || (ssidLen > 32)) {
        return FALSE;
    }

    passwordHash(password_p, ssid_p, ssidLen, keyText);
    memcpy(output_p, keyText, outputLen);

    return TRUE;
}
