/*
** $Id: rate.c 1.23 2004/10/04 18:09:51 cmchang Exp $
*/

/*************************************************************************
* Copyright (c) 2002 Inprocomm, Inc.
*
* All rights reserved. Copying, compilation, modification, distribution
* or any other use whatsoever of this material is strictly prohibited
* except in accordance with a Software License Agreement with
* Inprocomm, Inc.
**************************************************************************/
/*************************************************************************
GENERAL DESCRIPTION
********************
This file contains the transmission rate handling routines for INPROCOMM
802.11 Wireless LAN Adapters.
*/

/************************************************************************
*                E X T E R N A L   R E F E R E N C E S
*************************************************************************
*/
#include "precomp.h"

/************************************************************************
*                          C O N S T A N T S
*************************************************************************
*/
#define VALID_DATA_RATE_LIST_LEN \
        (sizeof(validDataRateList) / sizeof(DATA_RATE_INFO))

/************************************************************************
*                         D A T A   T Y P E S
*************************************************************************
*/

/************************************************************************
*                        P U B L I C   D A T A
*************************************************************************
*/

/************************************************************************
*                       P R I V A T E   D A T A
*************************************************************************
*/
DEBUGFILE("rate.c");

static DATA_RATE_INFO validDataRateList[] = {
    { RATE_1M,   "1M",   RATE_SET_BIT_1M,   RATE_CODE_1M,   PHY_MODULATION_DSSS_CCK | PHY_PREAMBLE_LONG },
    { RATE_2M,   "2M",   RATE_SET_BIT_2M,   RATE_CODE_2M,   PHY_MODULATION_DSSS_CCK | PHY_PREAMBLE_LONG | PHY_PREAMBLE_SHORT },
    { RATE_5_5M, "5.5M", RATE_SET_BIT_5_5M, RATE_CODE_5_5M, PHY_MODULATION_DSSS_CCK | PHY_PREAMBLE_LONG | PHY_PREAMBLE_SHORT },
    { RATE_11M,  "11M",  RATE_SET_BIT_11M,  RATE_CODE_11M,  PHY_MODULATION_DSSS_CCK | PHY_PREAMBLE_LONG | PHY_PREAMBLE_SHORT },
    { RATE_6M,   "6M",   RATE_SET_BIT_6M,   RATE_CODE_6M,   PHY_MODULATION_OFDM },
    { RATE_12M,  "12M",  RATE_SET_BIT_12M,  RATE_CODE_12M,  PHY_MODULATION_OFDM },
    { RATE_24M,  "24M",  RATE_SET_BIT_24M,  RATE_CODE_24M,  PHY_MODULATION_OFDM },
    { RATE_54M,  "54M",  RATE_SET_BIT_54M,  RATE_CODE_54M,  PHY_MODULATION_OFDM },
    { RATE_9M,   "9M",   RATE_SET_BIT_9M,   RATE_CODE_9M,   PHY_MODULATION_OFDM },
    { RATE_18M,  "18M",  RATE_SET_BIT_18M,  RATE_CODE_18M,  PHY_MODULATION_OFDM },
    { RATE_36M,  "36M",  RATE_SET_BIT_36M,  RATE_CODE_36M,  PHY_MODULATION_OFDM },
    { RATE_48M,  "48M",  RATE_SET_BIT_48M,  RATE_CODE_48M,  PHY_MODULATION_OFDM },
    { RATE_22M,  "22M",  RATE_SET_BIT_22M,  0,              PHY_MODULATION_PBCC },
    { RATE_33M,  "33M",  RATE_SET_BIT_33M,  0,              PHY_MODULATION_PBCC }
};


/************************************************************************
*                             M A C R O S
*************************************************************************
*/

/************************************************************************
*              F U N C T I O N   D E C L A R A T I O N S
*************************************************************************
*/

/************************************************************************
*                          F U N C T I O N S
*************************************************************************
*/

/*_________________________________________________________________________
**  rateGetDataRateInfo
**
**  descriptions: This routine is called to get the information of the
**                specified data rate.
**  parameters:
**      dataRate - The data rate in units of 500 kbit/s
**  result:
**      dataRateInfo_p - Pointer to the information structure of the
**                       specified data rate
**  return:
**      TRUE - The given rate is valid and its information is returned.
**      FALSE - The given rate is invalid. No information is returned.
**  note:
**__________________________________________________________________________*/
BOOLEAN
rateGetDataRateInfo(IN UINT_8 dataRate,
                    OUT PDATA_RATE_INFO *dataRateInfo_p)
{
    UINT i;

    *dataRateInfo_p = NULL;

    for (i = 0; i < VALID_DATA_RATE_LIST_LEN; i++) {
        if (dataRate == validDataRateList[i].rate) {
            *dataRateInfo_p = &validDataRateList[i];

            return TRUE;
        }
    }

    return FALSE;
}   /* rateGetDataRateInfo */

/*_________________________________________________________________________
**  rateGetRatesFromRateSet
**
**  descriptions: This routine is called to get the valid rates from the
**                given supported and basic rate sets.
**  parameters:
**      supportedRateSet - The supported rate set
**      basicRateSet - The basic rate set
**      maxNumOfRates - The maximum number of rates that can be put into
**                      the buffer prepared by the caller
**  result:
**      rates_p - Pointer to the buffer into which the routine will return
**                the existing rates
**      numOfRates_p - Pointer to the number of rates that are returned
**  return:         (none)
**  note:
**__________________________________________________________________________*/
VOID
rateGetRatesFromRateSet(IN UINT_32 supportedRateSet,
                        IN UINT_32 basicRateSet,
                        OUT PUINT_8 rates_p,
                        IN UINT maxNumOfRates,
                        OUT PUINT numOfRates_p)
{
    UINT i, j;

    j = 0;

    for (i = 0; i < VALID_DATA_RATE_LIST_LEN; i++) {
        if (supportedRateSet & validDataRateList[i].rateSetBit) {
            *(rates_p + j) = validDataRateList[i].rate;

            if (basicRateSet & validDataRateList[i].rateSetBit) {
                *(rates_p + j) |= RATE_BASIC_BIT;
            }

            j++;
            if (j >= maxNumOfRates) {
                break;
            }
        }
    }

    *numOfRates_p = j;
}   /* rateGetRatesFromRateSet */

/*_________________________________________________________________________
**  rateGetHighestRate
**
**  descriptions: This routine is called to get the highest rate that
**                belongs to the given rate set and does not exceed the
**                specified upper limit.
**  parameters:
**      rateSet - The rate set from which the highest rate is searched
**      upperLimit - The upper limit that the highest rate shall be less
**                   than or equal to
**  result:
**      highestRate_p - Pointer to the highest rate that meets the
**                      requirement. If no highest rate is found, a value
**                      of zero is returned.
**      phyMode_p - Pointer to the PHY mode to which the highest rate
**                  belongs. If it is set to NULL, we don't return the PHY
**                  mode.
**  return:
**      TRUE - The highest rate in the given rate set is found.
**      FALSE - No highest rate in the given rate set is found.
**  note:
**__________________________________________________________________________*/
BOOLEAN
rateGetHighestRate(IN UINT_32 rateSet,
                   IN UINT_8 upperLimit,
                   OUT PUINT_8 highestRate_p,
                   OUT PUINT_32 phyMode_p)
{
    if ((rateSet & RATE_SET_BIT_54M) && (upperLimit >= RATE_54M)) {
        *highestRate_p = RATE_54M;
        if (phyMode_p) {
            *phyMode_p = PHY_MODULATION_OFDM;
        }
        return TRUE;
    }

    if ((rateSet & RATE_SET_BIT_48M) && (upperLimit >= RATE_48M)) {
        *highestRate_p = RATE_48M;
        if (phyMode_p) {
            *phyMode_p = PHY_MODULATION_OFDM;
        }
        return TRUE;
    }

    if ((rateSet & RATE_SET_BIT_36M) && (upperLimit >= RATE_36M)) {
        *highestRate_p = RATE_36M;
        if (phyMode_p) {
            *phyMode_p = PHY_MODULATION_OFDM;
        }
        return TRUE;
    }

    if ((rateSet & RATE_SET_BIT_24M) && (upperLimit >= RATE_24M)) {
        *highestRate_p = RATE_24M;
        if (phyMode_p) {
            *phyMode_p = PHY_MODULATION_OFDM;
        }
        return TRUE;
    }

    if ((rateSet & RATE_SET_BIT_18M) && (upperLimit >= RATE_18M)) {
        *highestRate_p = RATE_18M;
        if (phyMode_p) {
            *phyMode_p = PHY_MODULATION_OFDM;
        }
        return TRUE;
    }

    if ((rateSet & RATE_SET_BIT_12M) && (upperLimit >= RATE_12M)) {
        *highestRate_p = RATE_12M;
        if (phyMode_p) {
            *phyMode_p = PHY_MODULATION_OFDM;
        }
        return TRUE;
    }

    if ((rateSet & RATE_SET_BIT_11M) && (upperLimit >= RATE_11M)) {
        *highestRate_p = RATE_11M;
        if (phyMode_p) {
            *phyMode_p = PHY_MODULATION_DSSS_CCK | PHY_PREAMBLE_LONG |
                PHY_PREAMBLE_SHORT;
        }
        return TRUE;
    }

    if ((rateSet & RATE_SET_BIT_9M) && (upperLimit >= RATE_9M)) {
        *highestRate_p = RATE_9M;
        if (phyMode_p) {
            *phyMode_p = PHY_MODULATION_OFDM;
        }
        return TRUE;
    }

    if ((rateSet & RATE_SET_BIT_6M) && (upperLimit >= RATE_6M)) {
        *highestRate_p = RATE_6M;
        if (phyMode_p) {
            *phyMode_p = PHY_MODULATION_OFDM;
        }
        return TRUE;
    }

    if ((rateSet & RATE_SET_BIT_5_5M) && (upperLimit >= RATE_5_5M)) {
        *highestRate_p = RATE_5_5M;
        if (phyMode_p) {
            *phyMode_p = PHY_MODULATION_DSSS_CCK | PHY_PREAMBLE_LONG |
                PHY_PREAMBLE_SHORT;
        }
        return TRUE;
    }

    if ((rateSet & RATE_SET_BIT_2M) && (upperLimit >= RATE_2M)) {
        *highestRate_p = RATE_2M;
        if (phyMode_p) {
            *phyMode_p = PHY_MODULATION_DSSS_CCK | PHY_PREAMBLE_LONG |
                PHY_PREAMBLE_SHORT;
        }
        return TRUE;
    }

    if ((rateSet & RATE_SET_BIT_1M) && (upperLimit >= RATE_1M)) {
        *highestRate_p = RATE_1M;
        if (phyMode_p) {
            *phyMode_p = PHY_MODULATION_DSSS_CCK | PHY_PREAMBLE_LONG;
        }
        return TRUE;
    }

    *highestRate_p = 0;

    if (phyMode_p) {
        *phyMode_p = 0;
    }

    return FALSE;
}   /* rateGetHighestRate */

/*_________________________________________________________________________
**  rateGetLowestRate
**
**  descriptions: This routine is called to get the lowest rate that
**                belongs to the given rate set and is not less than the
**                specified lower limit.
**  parameters:
**      rateSet - The rate set from which the lowest rate is searched
**      lowerLimit - The lower limit that the lowest rate shall be greater
**                   than or equal to
**  result:
**      lowestRate_p - Pointer to the highest rate that meets the
**                     requirement. If no lowest rate is found, a value
**                     of zero is returned.
**      phyMode_p - Pointer to the PHY mode to which the lowest rate
**                  belongs. If it is set to NULL, we don't return the PHY
**                  mode.
**  return:
**      TRUE - The lowest rate in the given rate set is found.
**      FALSE - No lowest rate in the given rate set is found.
**  note:
**__________________________________________________________________________*/
BOOLEAN
rateGetLowestRate(IN UINT_32 rateSet,
                  IN UINT_8 lowerLimit,
                  OUT PUINT_8 lowestRate_p,
                  OUT PUINT_32 phyMode_p)
{
    if ((rateSet & RATE_SET_BIT_1M) && (lowerLimit <= RATE_1M)) {
        *lowestRate_p = RATE_1M;
        if (phyMode_p) {
            *phyMode_p = PHY_MODULATION_DSSS_CCK | PHY_PREAMBLE_LONG;
        }
        return TRUE;
    }

    if ((rateSet & RATE_SET_BIT_2M) && (lowerLimit <= RATE_2M)) {
        *lowestRate_p = RATE_2M;
        if (phyMode_p) {
            *phyMode_p = PHY_MODULATION_DSSS_CCK | PHY_PREAMBLE_LONG |
                PHY_PREAMBLE_SHORT;
        }
        return TRUE;
    }

    if ((rateSet & RATE_SET_BIT_5_5M) && (lowerLimit <= RATE_5_5M)) {
        *lowestRate_p = RATE_5_5M;
        if (phyMode_p) {
            *phyMode_p = PHY_MODULATION_DSSS_CCK | PHY_PREAMBLE_LONG |
                PHY_PREAMBLE_SHORT;
        }
        return TRUE;
    }

    if ((rateSet & RATE_SET_BIT_6M) && (lowerLimit <= RATE_6M)) {
        *lowestRate_p = RATE_6M;
        if (phyMode_p) {
            *phyMode_p = PHY_MODULATION_OFDM;
        }
        return TRUE;
    }

    if ((rateSet & RATE_SET_BIT_9M) && (lowerLimit <= RATE_9M)) {
        *lowestRate_p = RATE_9M;
        if (phyMode_p) {
            *phyMode_p = PHY_MODULATION_OFDM;
        }
        return TRUE;
    }

    if ((rateSet & RATE_SET_BIT_11M) && (lowerLimit <= RATE_11M)) {
        *lowestRate_p = RATE_11M;
        if (phyMode_p) {
            *phyMode_p = PHY_MODULATION_DSSS_CCK | PHY_PREAMBLE_LONG |
                PHY_PREAMBLE_SHORT;
        }
        return TRUE;
    }

    if ((rateSet & RATE_SET_BIT_12M) && (lowerLimit <= RATE_12M)) {
        *lowestRate_p = RATE_12M;
        if (phyMode_p) {
            *phyMode_p = PHY_MODULATION_OFDM;
        }
        return TRUE;
    }

    if ((rateSet & RATE_SET_BIT_18M) && (lowerLimit <= RATE_18M)) {
        *lowestRate_p = RATE_18M;
        if (phyMode_p) {
            *phyMode_p = PHY_MODULATION_OFDM;
        }
        return TRUE;
    }

    if ((rateSet & RATE_SET_BIT_24M) && (lowerLimit <= RATE_24M)) {
        *lowestRate_p = RATE_24M;
        if (phyMode_p) {
            *phyMode_p = PHY_MODULATION_OFDM;
        }
        return TRUE;
    }

    if ((rateSet & RATE_SET_BIT_36M) && (lowerLimit <= RATE_36M)) {
        *lowestRate_p = RATE_36M;
        if (phyMode_p) {
            *phyMode_p = PHY_MODULATION_OFDM;
        }
        return TRUE;
    }

    if ((rateSet & RATE_SET_BIT_48M) && (lowerLimit <= RATE_48M)) {
        *lowestRate_p = RATE_48M;
        if (phyMode_p) {
            *phyMode_p = PHY_MODULATION_OFDM;
        }
        return TRUE;
    }

    if ((rateSet & RATE_SET_BIT_54M) && (lowerLimit <= RATE_54M)) {
        *lowestRate_p = RATE_54M;
        if (phyMode_p) {
            *phyMode_p = PHY_MODULATION_OFDM;
        }
        return TRUE;
    }

    *lowestRate_p = 0;

    if (phyMode_p) {
        *phyMode_p = 0;
    }

    return FALSE;
}   /* rateGetLowestRate */

/*_________________________________________________________________________
**  rateGetTxRateCode
**
**  descriptions: This routine is called to get the code of the given
**                transmission rate to fill the TFCB.
**  parameters:
**      rate - The transmission rate used to send a frame
**      shortPreambleUsed - Set to TRUE if the frame is transmitted by
**                          using the short PLCP preamble and header of the
**                          ERP-DSSS/CCK mode. Otherwise, set to FALSE.
**  result:
**      rateCode_p - Pointer to the code of the given transmission rate
**  return:
**      TRUE - The corresponding rate code is found.
**      FALSE - No corresponding rate codes are found.
**  note:
**__________________________________________________________________________*/
BOOLEAN
rateGetTxRateCode(IN UINT_8 rate,
                  IN BOOLEAN shortPreambleUsed,
                  OUT PUINT_32 rateCode_p)
{
    PDATA_RATE_INFO rateInfo_p;

    if (!rateGetDataRateInfo(rate, &rateInfo_p)) {
        return FALSE;
    }

    *rateCode_p = rateInfo_p->rateCode;

    if (rateInfo_p->phyMode & PHY_MODULATION_OFDM) {
        *rateCode_p |= RATE_CODE_OFDM;
    } else if (rateInfo_p->phyMode & PHY_MODULATION_DSSS_CCK) {
        if ((rateInfo_p->phyMode & PHY_PREAMBLE_SHORT) && shortPreambleUsed) {
            *rateCode_p |= RATE_CODE_CCK | RATE_CODE_SHORT_PREAMBLE;
        } else {
            *rateCode_p |= RATE_CODE_CCK | RATE_CODE_LONG_PREAMBLE;
        }
    }

    return TRUE;
}   /* rateGetTxRateCode */

/*_________________________________________________________________________
**  rateGetRateByRateCode
**
**  descriptions: This routine is called to get the mapping rate by a
**                given rate code.
**  parameters:
**      rateCode - The rate code used to find its mapping rate
**  result:
**      rate_p - Pointer to the mapping rate
**  return:
**      TRUE - The corresponding rate is found.
**      FALSE - No corresponding rates are found.
**  note:
**__________________________________________________________________________*/
BOOLEAN
rateGetRateByRateCode(IN UINT_32 rateCode,
                      OUT PUINT_8 rate_p)
{
    BOOLEAN status = TRUE;

    switch (rateCode & RATE_CODE_RATE_MASK) {
        case RATE_CODE_54M:
             *rate_p = RATE_54M;
             break;

        case RATE_CODE_48M:
             *rate_p = RATE_48M;
             break;

        case RATE_CODE_36M:
             *rate_p = RATE_36M;
             break;

        case RATE_CODE_24M:
             *rate_p = RATE_24M;
             break;

        case RATE_CODE_18M:
             *rate_p = RATE_18M;
             break;

        case RATE_CODE_12M:
             *rate_p = RATE_12M;
             break;

        case RATE_CODE_9M:
             *rate_p = RATE_9M;
             break;

        case RATE_CODE_6M:
             *rate_p = RATE_6M;
             break;

        case RATE_CODE_11M:
             *rate_p = RATE_11M;
             break;

        case RATE_CODE_5_5M:
             *rate_p = RATE_5_5M;
             break;

        case RATE_CODE_2M:
             *rate_p = RATE_2M;
             break;

        case RATE_CODE_1M:
             *rate_p = RATE_1M;
             break;

        default:
             *rate_p = 0;
             status = FALSE;
    }

    return status;
}   /* rateGetRateByRateCode */

/*_________________________________________________________________________
**  rateComputePpduDuration
**
**  descriptions: This routine is called to compute the PPDU duration of a
**                MPDU.
**  parameters:
**      mpduLen - The length of the MPDU, in number of octets
**      dataRate - The data rate used to transmit this MPDU
**      shortPreambleUsed - Set to TRUE if this MPDU is transmitted by
**                          using the short PLCP preamble and header of the
**                          ERP-DSSS/CCK mode. Otherwise, set to FALSE.
**  result:
**      ppduDuration_p - Pointer to the PPDU duration, in units of
**                       microseconds
**  return:         (none)
**  note:
**__________________________________________________________________________*/
VOID
rateComputePpduDuration(IN UINT_32 mpduLen,
                        IN UINT_8 dataRate,
                        IN BOOLEAN shortPreambleUsed,
                        OUT PUINT_16 ppduDuration_p)
{
    UINT_16 numSymbols;

    switch (dataRate) {
        case RATE_54M:
        case RATE_48M:
        case RATE_36M:
        case RATE_24M:
        case RATE_18M:
        case RATE_12M:
        case RATE_9M:
        case RATE_6M:
             numSymbols = (UINT_16) ((mpduLen * 4 + 10 + (UINT_32) dataRate) /
                 (UINT_32) dataRate);

             *ppduDuration_p = 16 + (1 + numSymbols) * 4 + 6;
             break;

        case RATE_1M:
             shortPreambleUsed = FALSE;

             /* ** Note: No "break" here. Go through the next case. */

        case RATE_11M:
        case RATE_5_5M:
        case RATE_2M:
             *ppduDuration_p = (UINT_16) ((mpduLen * 16 +
                 (UINT_32) dataRate - 1) / (UINT_32) dataRate);

             if (shortPreambleUsed) {
                 *ppduDuration_p += 96;
             } else {
                 *ppduDuration_p += 192;
             }
             break;

        default:
             *ppduDuration_p = 0;
    }
}   /* rateComputePpduDuration */

/*_________________________________________________________________________
**  rateComputeCtsAckDuration
**
**  descriptions: This routine is called to compute the PPDU duration of a
**                CTS or ACK frame.
**  parameters:
**      dataRate - The data rate used to transmit a CTS or ACK frame
**      shortPreambleUsed - Set to TRUE if this CTS or ACK frame is
**                          transmitted by using the short PLCP preamble
**                          and header of the ERP-DSSS/CCK mode.
**                          Otherwise, set to FALSE.
**  result:
**      ctsAckDuration_p - Pointer to the PPDU duration of a CTS or ACK
**                         frame, in units of microseconds
**  return:         (none)
**  note:
**__________________________________________________________________________*/
VOID
rateComputeCtsAckDuration(IN UINT_8 dataRate,
                          IN BOOLEAN shortPreambleUsed,
                          OUT PUINT_16 ctsAckDuration_p)
{
    switch (dataRate) {
        case RATE_54M:
        case RATE_48M:
        case RATE_36M:
             *ctsAckDuration_p = 30;
             break;

        case RATE_24M:
        case RATE_18M:
             *ctsAckDuration_p = 34;
             break;

        case RATE_12M:
             *ctsAckDuration_p = 38;
             break;

        case RATE_9M:
             *ctsAckDuration_p = 42;
             break;

        case RATE_6M:
             *ctsAckDuration_p = 50;
             break;

        case RATE_11M:
             if (shortPreambleUsed) {
                 *ctsAckDuration_p = 107;
             } else {
                 *ctsAckDuration_p = 203;
             }
             break;

        case RATE_5_5M:
             if (shortPreambleUsed) {
                 *ctsAckDuration_p = 117;
             } else {
                 *ctsAckDuration_p = 213;
             }
             break;

        case RATE_2M:
             if (shortPreambleUsed) {
                 *ctsAckDuration_p = 152;
             } else {
                 *ctsAckDuration_p = 248;
             }
             break;

        case RATE_1M:
             *ctsAckDuration_p = 304;
             break;

        default:
             *ctsAckDuration_p = 0;
    }
}   /* rateComputeCtsAckDuration */

/*_________________________________________________________________________
**  rateSetupSupportedRateInfo
**
**  descriptions: This routine is called to set up the BSS supported rate
**                information.
**  parameters:
**      supRatesArray_p - Pointer to the array of the BSS supported rate
**                        information structures
**      numSupRates - The number of BSS supported rates
**      bssBasicRateSet - The basic rate set of the BSS
**  result:
**  return:         (none)
**  note:
**__________________________________________________________________________*/
VOID
rateSetupSupportedRateInfo(IN OUT PSUPPORTED_RATE_INFO supRatesArray_p,
                           IN UINT numSupRates,
                           IN UINT_32 bssBasicRateSet)
{
    UINT i;
    UINT_8 rate;
    PDATA_RATE_INFO rateInfo_p;
    UINT_32 rateSetMask;
    UINT_8 ackRate;

    DEBUGFUNC("rateSetupSupportedRateInfo");

    for (i = 0; i < numSupRates; i++) {
        rate = supRatesArray_p[i].rate;

        if (!rateGetDataRateInfo(rate, &rateInfo_p)) {
            continue;
        }

        supRatesArray_p[i].phyMode = rateInfo_p->phyMode;

        if (rateInfo_p->phyMode & PHY_MODULATION_OFDM) {
            rateSetMask = RATE_SET_MASK_OFDM;
        } else {
            rateSetMask = RATE_SET_MASK_DSSS_CCK;
        }

        if (rateGetHighestRate(bssBasicRateSet & rateSetMask, rate, &ackRate,
            NULL)) {
            if (ackRate > RATE_36M && (bssBasicRateSet & RATE_SET_BIT_36M)) {
                ackRate = RATE_36M;
            }
        } else {
            if (rateInfo_p->phyMode & PHY_MODULATION_OFDM) {
                if (rate >= RATE_24M) {
                    ackRate = RATE_24M;
                } else if (rate >= RATE_12M) {
                    ackRate = RATE_12M;
                } else {
                    ackRate = RATE_6M;
                }
            } else {
                ackRate = rate;
            }
        }

        supRatesArray_p[i].ackTxRate = ackRate;

        rateComputeCtsAckDuration(ackRate, FALSE,
            &supRatesArray_p[i].txDuration);

        if (!rateGetHighestRate(bssBasicRateSet & RATE_SET_MASK_DSSS_CCK,
            rate, &supRatesArray_p[i].protectFrameTxRate,
            NULL)) {
            supRatesArray_p[i].protectFrameTxRate = RATE_1M;
        }

        DBGLOG(MGT, TRACE,
            ("Supported rate[%d]: %s -- ACK_rate=0x%02x (dur=%d), protect_rate=0x%02x\n",
            i, rateInfo_p->displayString, supRatesArray_p[i].ackTxRate,
            supRatesArray_p[i].txDuration,
            supRatesArray_p[i].protectFrameTxRate));
    }
}   /* rateSetupSupportedRateInfo */

/*_________________________________________________________________________
**  rateUpgradeTxRate
**
**  descriptions: This routine is called to upgrade the current
**                transmission rate.
**  parameters:
**      txRateSet - The transmission rate set for rate switching
**      currentTxRate - The current transmission rate
**  result:
**      nextRate_p - Pointer to the next rate to upgrade the Tx rate. If
**                   there are not any next rates to upgrade the Tx rate,
**                   a value of zero is returned.
**      phyMode_p - Pointer to the PHY mode to which the next rate
**                  belongs. If it is set to NULL, we don't return the PHY
**                  mode.
**  return:         (none)
**  note:
**__________________________________________________________________________*/
VOID
rateUpgradeTxRate(IN UINT_32 txRateSet,
                  IN UINT_8 currentTxRate,
                  OUT PUINT_8 nextRate_p,
                  OUT PUINT_32 phyMode_p)
{
    switch (currentTxRate) {
        case RATE_1M:
             if (txRateSet & RATE_SET_BIT_2M) {
                *nextRate_p = RATE_2M;
                if (phyMode_p) {
                    *phyMode_p = PHY_MODULATION_DSSS_CCK | PHY_PREAMBLE_LONG |
                        PHY_PREAMBLE_SHORT;
                }
                break;
             }

        case RATE_2M:
             if (txRateSet & RATE_SET_BIT_5_5M) {
                 *nextRate_p = RATE_5_5M;
                if (phyMode_p) {
                    *phyMode_p = PHY_MODULATION_DSSS_CCK | PHY_PREAMBLE_LONG |
                        PHY_PREAMBLE_SHORT;
                }
                break;
             }

             if (txRateSet & RATE_SET_BIT_6M) {
                 *nextRate_p = RATE_6M;
                if (phyMode_p) {
                    *phyMode_p = PHY_MODULATION_OFDM;
                }
                break;
             }

        case RATE_5_5M:
        case RATE_6M:
             if (txRateSet & RATE_SET_BIT_11M) {
                 *nextRate_p = RATE_11M;
                if (phyMode_p) {
                    *phyMode_p = PHY_MODULATION_DSSS_CCK | PHY_PREAMBLE_LONG |
                        PHY_PREAMBLE_SHORT;
                }
                break;
             }

             if (txRateSet & RATE_SET_BIT_9M) {
                 *nextRate_p = RATE_9M;
                if (phyMode_p) {
                    *phyMode_p = PHY_MODULATION_OFDM;
                }
                break;
             }

        case RATE_11M:
        case RATE_9M:
             if (txRateSet & RATE_SET_BIT_12M) {
                 *nextRate_p = RATE_12M;
                if (phyMode_p) {
                    *phyMode_p = PHY_MODULATION_OFDM;
                }
                break;
             }

        case RATE_12M:
             if (txRateSet & RATE_SET_BIT_18M) {
                 *nextRate_p = RATE_18M;
                if (phyMode_p) {
                    *phyMode_p = PHY_MODULATION_OFDM;
                }
                break;
             }

        case RATE_18M:
             if (txRateSet & RATE_SET_BIT_24M) {
                 *nextRate_p = RATE_24M;
                if (phyMode_p) {
                    *phyMode_p = PHY_MODULATION_OFDM;
                }
                break;
             }

        case RATE_24M:
             if (txRateSet & RATE_SET_BIT_36M) {
                 *nextRate_p = RATE_36M;
                if (phyMode_p) {
                    *phyMode_p = PHY_MODULATION_OFDM;
                }
                break;
             }

        case RATE_36M:
             if (txRateSet & RATE_SET_BIT_48M) {
                 *nextRate_p = RATE_48M;
                if (phyMode_p) {
                    *phyMode_p = PHY_MODULATION_OFDM;
                }
                break;
             }

        case RATE_48M:
             if (txRateSet & RATE_SET_BIT_54M) {
                 *nextRate_p = RATE_54M;
                if (phyMode_p) {
                    *phyMode_p = PHY_MODULATION_OFDM;
                }
                break;
             }

        default:
             *nextRate_p = 0;
    }
}   /* rateUpgradeTxRate */

/*_________________________________________________________________________
**  rateDegradeTxRate
**
**  descriptions: This routine is called to degrade the current
**                transmission rate.
**  parameters:
**      txRateSet - The transmission rate set for rate switching
**      currentTxRate - The current transmission rate
**  result:
**      nextRate_p - Pointer to the next rate to degrade the Tx rate. If
**                   there are not any next rates to degrade the Tx rate,
**                   a value of zero is returned.
**      phyMode_p - Pointer to the PHY mode to which the next rate
**                  belongs. If it is set to NULL, we don't return the PHY
**                  mode.
**  return:         (none)
**  note:
**__________________________________________________________________________*/
VOID
rateDegradeTxRate(IN UINT_32 txRateSet,
                  IN UINT_8 currentTxRate,
                  OUT PUINT_8 nextRate_p,
                  OUT PUINT_32 phyMode_p)
{
    switch (currentTxRate) {
        case RATE_54M:
             if (txRateSet & RATE_SET_BIT_48M) {
                *nextRate_p = RATE_48M;
                if (phyMode_p) {
                    *phyMode_p = PHY_MODULATION_OFDM;
                }
                break;
             }

        case RATE_48M:
             if (txRateSet & RATE_SET_BIT_36M) {
                 *nextRate_p = RATE_36M;
                if (phyMode_p) {
                    *phyMode_p = PHY_MODULATION_OFDM;
                }
                break;
             }

        case RATE_36M:
             if (txRateSet & RATE_SET_BIT_24M) {
                 *nextRate_p = RATE_24M;
                if (phyMode_p) {
                    *phyMode_p = PHY_MODULATION_OFDM;
                }
                break;
             }

        case RATE_24M:
             if (txRateSet & RATE_SET_BIT_18M) {
                 *nextRate_p = RATE_18M;
                if (phyMode_p) {
                    *phyMode_p = PHY_MODULATION_OFDM;
                }
                break;
             }

        case RATE_18M:
             if (txRateSet & RATE_SET_BIT_12M) {
                 *nextRate_p = RATE_12M;
                if (phyMode_p) {
                    *phyMode_p = PHY_MODULATION_OFDM;
                }
                break;
             }

        case RATE_12M:
             if (txRateSet & RATE_SET_BIT_11M) {
                 *nextRate_p = RATE_11M;
                if (phyMode_p) {
                    *phyMode_p = PHY_MODULATION_DSSS_CCK | PHY_PREAMBLE_LONG |
                        PHY_PREAMBLE_SHORT;
                }
                break;
             }

             if (txRateSet & RATE_SET_BIT_9M) {
                 *nextRate_p = RATE_9M;
                if (phyMode_p) {
                    *phyMode_p = PHY_MODULATION_OFDM;
                }
                break;
             }

        case RATE_9M:
        case RATE_11M:
             if (txRateSet & RATE_SET_BIT_5_5M) {
                 *nextRate_p = RATE_5_5M;
                if (phyMode_p) {
                    *phyMode_p = PHY_MODULATION_DSSS_CCK | PHY_PREAMBLE_LONG |
                        PHY_PREAMBLE_SHORT;
                }
                break;
             }

             if (txRateSet & RATE_SET_BIT_6M) {
                 *nextRate_p = RATE_6M;
                if (phyMode_p) {
                    *phyMode_p = PHY_MODULATION_OFDM;
                }
                break;
             }

        case RATE_6M:
        case RATE_5_5M:
             if (txRateSet & RATE_SET_BIT_2M) {
                 *nextRate_p = RATE_2M;
                if (phyMode_p) {
                    *phyMode_p = PHY_MODULATION_DSSS_CCK | PHY_PREAMBLE_LONG |
                        PHY_PREAMBLE_SHORT;
                }
                break;
             }

        case RATE_2M:
             if (txRateSet & RATE_SET_BIT_1M) {
                 *nextRate_p = RATE_1M;
                if (phyMode_p) {
                    *phyMode_p = PHY_MODULATION_DSSS_CCK | PHY_PREAMBLE_LONG;
                }
                break;
             }

        default:
             *nextRate_p = 0;
    }
}   /* rateDegradeTxRate */

